/* Example DevMUD module to exercise interface standard and socket module */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include "../include/devmud.h"

#define SNEAKY_FLAG 1

/* TODO make player struct dynamic, check for more players ok in select */
#define MAX_PLAYERS 200
#define MAX_MOBS 4
#define READ_BUFFER_SIZE 16384
#define OUT_BUFFER_SIZE (READ_BUFFER_SIZE+256)
#define LASTLOG_SIZE 10

static void (*queue_event)(struct timeval when, void (*func)(struct timeval when, int data), int data) = NULL;
static void (*close_connection)(int which) = NULL;
static int (*real_connection)(int which) = NULL;
static char *(*connection_name)(int which) = NULL;
static struct timeval (*connection_time)(int which) = NULL;
static int (*open_port)
		(int port,
		 unsigned int addr,
		 int (*new_connection)
		       (int id,
			void (*message)(int id,char *message,int len,int oob)),
		 void (*end_connection)(int id),
		 void (*message)(int id,char *message,int len,int oob))
	= NULL;
static void (*close_port)(int id) = NULL;

/* TODO cleanup state on player exit to mark as avail */

struct playerinfo {
	/* function to send messages */
	void (*message)(int id,char *message,int len,int oob);
	int id;	/* network ID */
	char command_buffer[READ_BUFFER_SIZE];
	int command_tail, state;
};

static struct playerinfo p[MAX_PLAYERS];

static int port_number = 8080;

static void send_string(int playerid, char *str)
{
	p[playerid].message(p[playerid].id,str,-1,0);
}

static int new_player(int id,void (*message)(int id,char *message,int len,int oob))
{
	int playerid;

	/* find a playerid */
	for (playerid=0;
	     (playerid<MAX_PLAYERS) && (p[playerid].state != -1);
	     playerid++)
		;	/* do nothing */
	if (playerid >= MAX_PLAYERS)
	{
		/* TODO bail on player */
	}
	
	p[playerid].state = 0;
	p[playerid].message = message;
	p[playerid].id = id;
	return(playerid);	/* what ID to use to contact us */
}

static void parse_input(int playerid, char *buffer, int len, int oob)
{
	char *command;

	if (oob)
	{
		return;
	}
	/* read until newline, or... */
	while ((*buffer != '\n') && (*buffer != '\r') &&
	/* we really filled up the buffer. */
	       (p[playerid].command_tail < READ_BUFFER_SIZE))
	{
		p[playerid].command_buffer[p[playerid].command_tail] = *buffer;
		if (*buffer == 0)
		{
			return;
		}
		p[playerid].command_tail++;
		buffer++;
	}
	/* terminate the string, if needed */
	p[playerid].command_buffer[p[playerid].command_tail] = 0;
	/* and inform the next time that there isn't a partial buffer */
	p[playerid].command_tail = 0;
	command = p[playerid].command_buffer;

	fprintf(stderr, "%s\n", command);
	send_string(playerid, "<HTML><HEAD><TITLE>Test Server</TITLE></HEAD><BODY>No real content</BODY></HTML>\n");
/*
	close_connection(p[playerid].id);
*/
}

static void end_player(int playerid)
{
	/* reinit player slot */
	p[playerid].command_buffer[0] = 0;
	p[playerid].command_tail = 0;
	p[playerid].state = -1;	/* open slot */
}

static void config_port(int port) 
{
	port_number = port;
}

static void config_opt(char *opt, char *val) 
{
	long int port;
	char *endptr;

	if (!strcmp(opt,"port"))
	{
		port = strtol(val,&endptr,0);
		/* converted all of val? */
		if ((*endptr == 0) && (endptr != val))
		{
			port_number = port;
		}
	}
}

struct interface web_supplies[] = {
	{ "config_port", config_port, "void(int)", NULL },
	{ "config_opt", config_opt, "void(char*,char*)", NULL },
	{NULL,NULL,NULL,NULL}
};

struct interface web_uses[] = {
	{ "queue_event", NULL, "void(timeval,void(*)(timeval,int),int)", NULL},
	{ "close_connection", NULL, "void(int)", NULL },
	{ "real_connection", NULL, "int(int)", NULL },
	{ "connection_name", NULL, "char*(int)", NULL },
	{ "connection_time", NULL, "timeval(int)", NULL },
	{ "open_port", NULL, "int(int,unsigned int,int(*)(int,void(*)(int,char*,int,int)),void(*)(int),void(*)(int,char*,int,int))", NULL },
	{ "close_port", NULL, "void(int)", NULL },
	{ NULL, NULL, NULL, NULL }
};

void web_use_functions(struct interface *it)
{
	struct interface *temp;

	temp = it;
	while (temp->name != NULL)
	{
		if ((!strcmp(temp->name,"queue_event")) &&
		    (!strcmp(temp->prototype,"void(timeval,void(timeval,int),int)")))
		{
			queue_event = temp->function;
		}
		else if ((!strcmp(temp->name,"close_connection")) &&
			 (!strcmp(temp->prototype,"void(int)")))
		{
			close_connection = temp->function;
		}
		else if ((!strcmp(temp->name,"real_connection")) &&
			 (!strcmp(temp->prototype,"int(int)")))
		{
			real_connection = temp->function;
		}
		else if ((!strcmp(temp->name,"connection_name")) &&
			 (!strcmp(temp->prototype,"char*(int)")))
		{
			connection_name = temp->function;
		}
		else if ((!strcmp(temp->name,"connection_time")) &&
			 (!strcmp(temp->prototype,"timeval(int)")))
		{
			connection_time = temp->function;
		}
		else if ((!strcmp(temp->name,"open_port")) &&
			 (!strcmp(temp->prototype,"int(int,unsigned int,int(*)(int,void(*)(int,char*,int,int)),void(*)(int),void(*)(int,char*,int,int))")))
		{
			open_port = temp->function;
		}
		else if ((!strcmp(temp->name,"close_port")) &&
			 (!strcmp(temp->prototype,"void(int)")))
		{
			close_port = temp->function;
		}
		temp++;
	}
	return;
}

int web_initialize(void)
{
	return(0);
}

/* TODO: move some of this to initialize... */
/* TODO: add support for a go function */
/* TODO: split into modules */
int web_start(void)
{
	int i;

	if ((queue_event == NULL) ||
	    (close_connection == NULL) ||
	    (real_connection == NULL) ||
	    (connection_name == NULL) ||
	    (connection_time == NULL) ||
	    (open_port == NULL) ||
	    (close_port == NULL))
	{
		fprintf(stderr,"Needed functions still uninitialized in web_start!\n");
		return(-1);
	}

	open_port(port_number,INADDR_ANY,new_player,end_player,parse_input);

	gettimeofday(&current_time,NULL);

	/* initialize player data to known values */
	for(i=0;i<MAX_PLAYERS;i++)
	{
		p[i].command_buffer[0] = 0;
		p[i].command_tail = 0;
		p[i].state = -1;	/* open slot */
	}
	return(0);
}

int web_stop(void)
{
	return(-1);
}
